#!/usr/bin/env python
"""
Movies v0.3

Made by Florentin Sardan
florentinwww (at) gmail.com

Project's page:
  http://www.betterprogramming.com/filter-your-video-collection-with-python-and-imdbpy.html
My portfolio:
  http://www.betterprogramming.com/

Before using this script, make sure you have the required libraries

sudo pip install imdbpy
sudo pip install prettytable

sudo ln -s /path/to/movies.py /usr/local/bin/movies.py
OR
alias movies="/work/python/tools/movies.py"

scan and refresh movies from inside the path "/home/elmo/Movies". In case you don't use "-p", the default path is stored in DEFAULT_MOVIES_PATH (see below)
movies -p "/home/elmo/Movies" -s -r

movies with andy serkis
movies -l -a "+andy serkis"

movies with at least one of these actors playing
movies -l -a "?andy serkis,?angelina jolie"

movies with at least one favorite actor playing
movies -l -o actor

comedy movies but no dramas
movies -l -g "+comedy,-drama"

movies made by "lionsgate" production company
movies -c "lionsgate"

movies made by any of the favorite production companies
movies -o company
"""

from optparse import OptionParser
import os, sys, re
import pickle
from pprint import pprint
import prettytable
from imdb import IMDb

INCLUDE = '+'
EXCLUDE = '-'
MAYBE = '?'
IMDB_CACHE_FILE = 'imdb.info'
IMDB_URL_FILE = 'imdb.html'
DEFAULT_MOVIES_PATH = '/home/elmo/Movies'
MOVIE_EXTENSIONS = ['avi', 'mkv', 'mpeg']

companies = [
'UTV Motion Pictures',
'Warner Bros. Pictures',
'Lionsgate',
'Twentieth Century Fox Film Corporation',
'Columbia Pictures',
'Paramount Pictures',
'Universal Pictures',
'Metro-Goldwyn-Mayer (MGM)',
'Walt Disney Pictures',
'Ilion Animation',
]
companies = map(str.lower, companies)

""" a list of good actors, i'll be using it to predefined some filters """
actors = """
Ben Stiller
Adam Sandler
Adrien Brody
Forest Whitaker
Jonathan Rhys Meyers
Ben Kingsley
Joaquin Phoenix
Scarlett Johansson
Jennifer Aniston
Jessica Alba
1. Robert De Niro
2. Tom Hanks
3. Al Pacino
4. Johnny Depp
5. Marlon Brando
6. Jack Nicholson
7. Meryl Streep
8. Leonardo DiCaprio
9. Morgan Freeman
10. Denzel Washington
11. Dustin Hoffman
12. Anthony Hopkins
13. Russell Crowe
14. Will Smith
15. Daniel Day Lewis
16. Clint Eastwood
17. Brad Pitt
18. Sean Penn
19. James Stewart
20. Paul Newman
21. Humphrey Bogart
22. Heath Ledger
23. Edward Norton
24. Kate Winslet
25. Tom Cruise
26. Harrison Ford
27. Audrey Hepburn
28. Mel Gibson
29. Kevin Spacey
30. Bette Davis
31. Cary Grant
32. Katharine Hepburn
33. Jim Carrey
34. Matt Damon
35. Bruce Willis
36. Sean Connery
37. Robert Downey Jr.
38. Julia Roberts
39. Christian Bale
40. George Clooney
41. Laurence Olivier
42. Robin Williams
43. John Wayne
44. Samuel L Jackson
45. Spencer Tracy
46. Cate Blanchett
47. Gregory Peck
48. Gene Hackman
49. Robert Duvall
50. Clark Gable
51. Jodie Foster
52. Nicolas Cage
53. Henry Fonda
54. Charlie Chaplin
55. Ingrid Bergman
56. Sandra Bullock
57. Gary Cooper
58. Sylvester Stallone
59. Nicole Kidman
60. John Travolta
61. Angelina Jolie
62. Robert Redford
63. James Dean
64. Ian McKellen
65. Jack Lemmon
66. Vivien Leigh
67. Judi Dench
68. Charlton Heston
69. Christopher Walken
70. Julie Andrews
71. Gary Oldman
72. Hugh Jackman
73. Elizabeth Taylor
74. Michael Caine
75. Steve McQueen
76. Charlize Theron
77. Reese Witherspoon
78. James Cagney
79. Adam Sandler
80. Phillip Seymour Hoffman
81. Toshiro Mifune
82. Shia LaBeouf
83. Peter O'Toole
84. Alec Guinness
85. Orlando Bloom
86. Viggo Mortensen
87. Sidney Poitier
88. Helena Bonham Carter
89. Grace Kelly
90. Diane Keaton
91. Orson Welles
92. Liam Neeson
93. Kevin Costner
94. Emma Thompson
95. Joe Pesci
96. Keira Knightley
97. Burt Lancaster
98. Halle Berry
99. Jeremy Irons
100. Harvey Keitel
"""
actors = map(str.lower, list(set([re.sub(r'^\s*\d+\.?\s*', '', actor) for actor in actors.split("\n")])))
actors.remove('')

def parse_options():
    usage = "usage: %prog [options]"
    parser = OptionParser(usage, version="%prog 1.0")
    
    parser.set_defaults(scan=False, refresh=False,
                        verbose=True, list=True, genres='', actors='', 
                        path=DEFAULT_MOVIES_PATH, 
                        )
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose")
    parser.add_option("-q", "--quiet",
                      action="store_false", dest="verbose")
    
    parser.add_option("-p", "--path", dest="path",
                      help="movie source, defaults to current directory")
    
    parser.add_option("-s", "--scan",
                      action="store_true", dest="scan",
                      help="scan the current directory for movies and write the IMDB info (ignore movies with existing data)")
    
    parser.add_option("-r", "--refresh",
                      action="store_true", dest="refresh",
                      help="scan the current directory for movies and refresh the IMDB info")
    
    parser.add_option("-l", "--list",
                      action="store_true", dest="list",
                      help="list movie information")

    parser.add_option("-g", "--genres",
                      action="store", dest="genres", default="",
                      help="filter by genres")

    parser.add_option("-a", "--actors",
                      action="store", dest="actors", default="",
                      help="filter by actors")
    
    parser.add_option("-c", "--companies",
                      action="store", dest="companies", default="",
                      help="filter by production companies")
    
    parser.add_option("-i", "--info",
                      action="store", dest="info", default="",
                      help="information about a movie identified by the imdb id")
    
    parser.add_option("-o", "--opt",
                      action="store", dest="opt", default="",
                      help="filter by predefined options")
    return parser

def main():
    parser = parse_options()

    (options, args) = parser.parse_args()

    if options.scan: # default is False
        movies = scan_movies(options.path)
        write_imdb_info(movies, options.refresh)

    """ prepare the filters if case """    
    filters = set_filters({}, 'genres', options.genres)
    filters = set_filters(filters, 'actors', options.actors)
    filters = set_filters(filters, 'companies', options.companies)
    
    """ set predefined filters """
    favorites = {
                 'actor': {'actors':{MAYBE: actors}},
                 'comedy': {'genres': {INCLUDE: ['comedy'], '-': ['drama']}},
                 'company': {'companies': {MAYBE: companies}},
                 }
    if options.opt:
        filters.update(favorites[options.opt])
    
    if options.info:
        options.list = False
        movie_info(options.info)
    
    if options.list: # default True
        movies = scan_movies(options.path)
        filtered_movies = filter_movies(movies, filters)
        
        x = prettytable.PrettyTable(["Imdb ID", "IMDB Movie", "Genres", "Rate", "Movie Dir Name"])
        x.align = "l"
        #x.hrules = prettytable.ALL
        x.set_field_align("IMDB Movie", "l") # Left align city names
        for key, movie in enumerate(filtered_movies):
            genres = ','.join((genre[0:4] for genre in movie.get('genres').split(',')))
            x.add_row([movie.get('imdb').movieID, movie.get('title')[0:28], genres, movie.get('rating'), os.path.basename(movie.get('imdb')['path'])[:75]])

        #x.printt(sortby="Rating", reversesort=True)
        x.printt()
        #x.print_html()

    #if len(args) != 1:
    #    parser.error("incorrect number of arguments")
    #if options.verbose:
    #    print "reading %s..." % options.filename
    return 0 # success

def movie_info(imdb_id):
    movie = find_imdb_movie(None, imdb_id)
    if not movie.get('title'):
        print "Movie %s not found."%imdb_id
        return False
    else:
        pprint(vars(movie))
    #pprint(vars(m['production companies'][0]))
    #sys.exit()
    #pprint(str(m['production companies'][0]))

def set_filters(filters, what, filter_string=None):
    """
    movie filters set by user
    example:
    filters = {'genres':{'+':['comedy','drama']}}
    possible filter keys: genres, actors
    possbile dict keys: +,-,?
    + = movie attribute (ie genre, actors) must include the all parameters
    - = movie attribute should not include any of the parameters
    ? = at least one of the parameters must exist. Useful when you search for movies having at least one favorite actor.
    """
    filters.setdefault(what, {})
    if not filter_string:
        return filters
    for option in filter_string.split(','):
        if option[0] in [INCLUDE, EXCLUDE, MAYBE]:
            filter_type = option[0]
            filter_value = option[1:].lower()
        else: # default filter type if none is defined
            filter_type = INCLUDE
            filter_value = option.lower()
        filters[what].setdefault(filter_type, []).append(filter_value)
    return filters

def movies_paths_to_dict(movies):
    """
    take a list of movie paths and return a list of movie dictionaries
    """
    pretty_movie_list = []
    for movie_path in movies:
        raw_movie_name = os.path.basename(movie_path)
        movie_name = clean_movie_name(raw_movie_name)
        imdb_id = get_imdb_id(raw_movie_name)
        imdb_file = os.path.join(movie_path, IMDB_CACHE_FILE)
        imdb_file_exists = os.path.isfile(imdb_file)
        pretty_movie_list.append({'path':movie_path,
                                  'name':movie_name,
                                  'imdb_file':imdb_file,
                                  'imdb_id':imdb_id,
                                  'imdb_file_exists':imdb_file_exists,
                                  })
    return pretty_movie_list

def filter_movies(movies, filters=None):
    """
    filter a movie list
    """
    filtered_movies = []
    pretty_movies_list = movies_paths_to_dict(movies)
    
    for movie in pretty_movies_list:
        if not movie['imdb_file_exists']:
            continue
        imdb_movie = pickle.load(open(movie['imdb_file'], 'r'))
        imdb_filters = {}
        imdb_filters['genres'] = [str(x).lower() for x in imdb_movie.get('genres', [])]
        imdb_filters['actors'] = [str(x).lower() for x in imdb_movie.get('cast', [])]
        imdb_filters['companies'] = [str(x).lower() for x in imdb_movie.get('production company', [])]
        
        try:
            filter_fail = []
            for filter_name, filter_dict in filters.items():
                if filter_dict.get(INCLUDE):
                    filter_fail.append( list(set(filter_dict.get(INCLUDE)) - set(imdb_filters.get(filter_name, []))) )
                if filter_dict.get(EXCLUDE):
                    filter_fail.append( list(set(filter_dict.get(EXCLUDE)) & set(imdb_filters.get(filter_name, []))) )
                if filter_dict.get(MAYBE):
                    filter_fail.append( not len(list(set(filter_dict.get(MAYBE)) & set(imdb_filters.get(filter_name, [])))) )
            if any(filter_fail):
                raise Exception('filter problem')
        except Exception as ex:
            continue
        else:
            movie['imdb'] = imdb_movie
            movie['title'] = imdb_movie.get('title') # movie['name']
            movie['rating'] = imdb_movie.get('rating')
            movie['genres'] = ",".join(imdb_filters['genres'])
            movie['actors'] = ",".join(imdb_filters['actors'])
            filtered_movies.append(movie)
    # sort by rating in reverse order (high ratings first)
    res = sorted(filtered_movies, key=lambda k: k.get('rating'), reverse=True)
    return res

def write_imdb_info(movies, refresh):
    """
    write imdb data to each movie directory
    """
    pretty_movies_list = movies_paths_to_dict(movies)
    for movie in pretty_movies_list:
        if (movie['imdb_file_exists'] and refresh) or not movie['imdb_file_exists']:
            print 'Searching imdb info: %s'%movie['name']
            imdb_movie = find_imdb_movie(movie['name'], movie['imdb_id'])
            if imdb_movie:
                cache_imdb(imdb_movie, movie['imdb_file'])

def cache_imdb(movie, imdb_file):
    movie['path'] = os.path.dirname(imdb_file)
    f = open(imdb_file, "w" )
    pickle.dump(movie, f)
    f.close()
    if IMDB_URL_FILE:
        url_file = os.path.join(movie['path'], IMDB_URL_FILE)
        f = open(url_file, "w" )
        f.write("""<html><head><meta http-equiv="REFRESH" content="0;url=%s"></head></html>"""%("http://www.imdb.com/title/tt%s/"%movie.movieID,))
        f.close()

def find_imdb_movie(movie_name, movie_imdb_id=None):
    """
    find a movie by name or imdb_id
    """
    imdb_movie = False
    ia = IMDb()
    if movie_imdb_id:
        imdb_movie = ia.get_movie(movie_imdb_id)
    else:
        movie_results = ia.search_movie(movie_name)
        if movie_results:
            imdb_movie = movie_results[0] # default chosen movie info is the first result
            for movie_res in movie_results:
                # find a result which matches the name of the movie we are looking for
                if movie_res['title'].lower()==movie_name.lower():
                    imdb_movie = movie_res
                    break
    if not imdb_movie:
            print 'WARNING: no IMDB info for "%s" (id:%s) was found'%(movie_name,str(movie_imdb_id))
            return False
    ia.update(imdb_movie)
    print 'Found movie: %s, %d'%(imdb_movie['title'], imdb_movie['year'])
    return imdb_movie

def scan_movies(path):
    """
    path is the absolute path where all the movies are located
    
    If you have several Movie directories on separate drives you can create one master Movie dir
    and symlink the other dirs inside it.
    
    Return a list of movie paths.
    The movie name is given by the directory name (not the movie file name)
    """

    movies = [] # is a list of paths to the directory name of the movie
    movie_generator = os.walk(os.path.abspath(path), True, None, True)
    for dirpath, dirnames, filenames in movie_generator:
        # first movie file registers the movie, skip the other movie files in the same directory
        if movies and movies[-1] in dirpath:
            # if the current directory is actually a subdirectory of a registered movie, skip it 
            continue
        for filename in filenames:
            if filetype(filename) in MOVIE_EXTENSIONS:
                movies.append(dirpath)
                break
    
    if not movies:
        print "No movies were found.\nMake sure the movie path is correct '%s' and there are movie directories and files there.\nTry running the script with '-s' or '--scan' argument"%path
        sys.exit()
    else:
        print "Processing %d movies..."%len(movies)
    return movies

def clean_movie_name(name):
    """
    clean a movie name
    TODO: rewrite the whole function
    """
    def stop_year(word):
        return re.search(r'^[12]\d{3}$', word)
    def stop_words(word):
        # case sensitive words
        c_words = ['LIMITED', 'LiMiTED', 'UNRATED', 'R5', 'DVDRip', 'PROPER']
        if word in c_words:
            return True
        
        # no case sensitive words
        nc_words = ['dvdrip', 'brrip', 'bdrip', '720p', '480p', 'klaxxon']
        if word.lower() in nc_words:
            return True

        return False
    stop_it = lambda word: stop_year(word) or stop_words(word)
    
    def ignore_words(word):
        return re.search(r'(?:tt)?(\d{7})', word)
    ignore_it = lambda word: ignore_words(word)
    
    good_name = []
    name = name.replace('.', ' ')
    words = re.split('\s|(?:\[|\()(\w+)(?:\]|\))', name)
    words = filter(lambda x:x, words)
    good_name.append(words.pop(0))

    for word in words:
        if stop_it(word):
            break
        else:
            if not ignore_it(word):
                good_name.append(word)
    return " ".join(good_name)

def get_imdb_id(name):
    match = re.search(r'(?:tt)?(\d{7})', name)
    return match.group(1) if match else None

def filetype(file):
    return file.split(".")[-1]

if __name__ == "__main__":
    status = main()
    sys.exit(status)
